#include <J3ML/Geometry/Triangle2D.hpp>
#include <J3ML/LinearAlgebra/Vector2.hpp>
#include <J3ML/LinearAlgebra/Matrix3x3.hpp>
#include <cfloat>

namespace J3ML::Geometry
{

    Triangle2D::Triangle2D(const Vector2& a, const Vector2& b, const Vector2& c) : A(a), B(b), C(c) {}

    void Triangle2D::Translate(const Vector2& offset)
    {
        A += offset;
        B += offset;
        C += offset;
    }

    void Triangle2D::Transform(const Matrix3x3& transform)
    {
        A = Mul2D(transform, A);
        B = Mul2D(transform, B);
        C = Mul2D(transform, C);
    }

    void Triangle2D::Transform(const Matrix4x4& transform)
    {
        A = MulPos2D(transform, A);
        B = MulPos2D(transform, B);
        C = MulPos2D(transform, C);
    }

    /// Implementation from Crister Ericson's Real-Time Collision Detection, pp. 51-52.
    inline float TriArea2D(float x1, float y1, float x2, float y2, float x3, float y3)
    {
        return (x1-x2)*(y2-y3) - (x2-x3)*(y1-y2);
    }

    Vector3 Triangle2D::BarycentricUVW(const Vector2& point) const
    {
        // Implementation from Crister Ericson's Real-Time Collision Detection, pp. 51-52, adapted for 2D.
        float nu, nv;

        nu = TriArea2D(point.x, point.y, B.x, B.y, C.x, C.y);
        nv = TriArea2D(point.x, point.y, C.x, C.y, A.x, A.y);
        float u = nu;
        float v = nv;
        float w = 1.f - u - v;
        return Vector3(u,v,w);
    }

    Vector2 Triangle2D::BarycentricUV(const Vector2& point) const
    {
        Vector3 uvw = BarycentricUVW(point);
        return Vector2(uvw.y, uvw.z);
    }

    bool Triangle2D::BarycentricInsideTriangle(const Vector3& barycentric)
    {
        return barycentric.x >= 0.f && barycentric.y >= 0 && barycentric.z >= 0.f &&
            Math::EqualAbs(barycentric.x + barycentric.y + barycentric.z, 1.f);
    }

    Vector2 Triangle2D::Point(float u, float v) const
    {
        // In case the triangle is far away from the origin but is small, the elements of 'a' will have large magnitudes,
        // and the elements of (b-a) and (c-a) will be much smaller quantities. Therefore, be extra careful with the
        // parenthesis and first sum the small floats together before adding it to the large one.
        return A + ((B-A) * u + (C-A) * v);
    }

    Vector2 Triangle2D::Point(float u, float v, float w) const
    {
        return u * A + v * B + w * C;
    }

    Vector2 Triangle2D::Point(const Vector3 &uvw) const
    {
        return Point(uvw.x, uvw.y, uvw.z);
    }

    Vector2 Triangle2D::Point(const Vector2 &uv) const
    {
        return Point(uv.x, uv.y);
    }

    Vector2 Triangle2D::Centroid() const
    {
        return (A + B + C) * (1.f/3.f);
    }

    float Triangle2D::Perimeter() const
    {
        return A.Distance(B) + B.Distance(C) + C.Distance(A);
    }

    Vector2 Triangle2D::Vertex(int i) const
    {
        assert(0 <= i);
        assert(i <= 2);
        if (i == 0)
            return A;
        else if (i == 1)
            return B;
        else if (i == 2)
            return C;
        else
            return Vector2::NaN;
    }


    Vector2 Triangle2D::ExtremePoint(const Vector2& direction) const
    {
        Vector2 mostExtreme = Vector2::NaN;
        float mostExtremeDist = -FLT_MAX;
        for (int i = 0; i < 3; ++i)
        {
            Vector2 pt = Vertex(i);
            float d = Vector2::Dot(direction, pt);
            if (d > mostExtremeDist)
            {
                mostExtremeDist = d;
                mostExtreme = pt;
            }
        }
        return mostExtreme;
    }

    //TODO must return on MSVC
    /*
    Vector2 Triangle2D::ExtremePoint(const Vector2& direction, float& projectionDistance) const
    {
        Vector2 extremePoint = ExtremePoint(direction);
        projectionDistance = extremePoint.Dot(direction);
    }
    */
}