#include <J3ML/LinearAlgebra/Matrix3x3.hpp>
#include <J3ML/LinearAlgebra/Matrix4x4.hpp>
#include <J3ML/LinearAlgebra/Matrices.inl>
#include <J3ML/LinearAlgebra/Vector3.hpp>
#include <J3ML/LinearAlgebra/Vector4.hpp>
#include <cmath>

namespace J3ML::LinearAlgebra {

    const Matrix3x3 Matrix3x3::Zero = Matrix3x3(0, 0, 0, 0, 0, 0, 0, 0, 0);
    const Matrix3x3 Matrix3x3::Identity = Matrix3x3(1, 0, 0, 0, 1, 0, 0, 0, 1);
    const Matrix3x3 Matrix3x3::NaN = Matrix3x3(NAN);

    Matrix3x3::Matrix3x3(float m00, float m01, float m02, float m10, float m11, float m12, float m20, float m21,
                         float m22) {
        this->elems[0][0] = m00;
        this->elems[0][1] = m01;
        this->elems[0][2] = m02;

        this->elems[1][0] = m10;
        this->elems[1][1] = m11;
        this->elems[1][2] = m12;

        this->elems[2][0] = m20;
        this->elems[2][1] = m21;
        this->elems[2][2] = m22;
    }

    Vector3 Matrix3x3::GetRow(int index) const {
        float x = this->elems[index][0];
        float y = this->elems[index][1];
        float z = this->elems[index][2];
        return {x, y, z};
    }

    Vector3 Matrix3x3::GetColumn(int index) const {
        float x = this->elems[0][index];
        float y = this->elems[1][index];
        float z = this->elems[2][index];

        return {x, y, z};
    }


    float Matrix3x3::At(int x, int y) const {
        return this->elems[x][y];
    }

    void Matrix3x3::SetAt(int x, int y, float value)
    {
        this->elems[x][y] = value;
    }

    Vector3 Matrix3x3::operator*(const Vector3 &rhs) const {
        return {
                At(0, 0) * rhs.x + At(0, 1) * rhs.y + At(0, 2) * rhs.z,
                At(1, 0) * rhs.x + At(1, 1) * rhs.y + At(1, 2) * rhs.z,
                At(2, 0) * rhs.x + At(2, 1) * rhs.y + At(2, 2) * rhs.z
        };
    }

    Matrix3x3 Matrix3x3::operator*(const Matrix3x3 &rhs) const {
        //Matrix3x3 r;
        auto m00 = At(0, 0) * rhs.At(0, 0) + At(0, 1) * rhs.At(1, 0) + At(0, 2) * rhs.At(2, 0);
        auto m01 = At(0, 0) * rhs.At(0, 1) + At(0, 1) * rhs.At(1, 1) + At(0, 2) * rhs.At(2, 1);
        auto m02 = At(0, 0) * rhs.At(0, 2) + At(0, 1) * rhs.At(1, 2) + At(0, 2) * rhs.At(2, 2);

        auto m10 = At(1, 0) * rhs.At(0, 0) + At(1, 1) * rhs.At(1, 0) + At(1, 2) * rhs.At(2, 0);
        auto m11 = At(1, 0) * rhs.At(0, 1) + At(1, 1) * rhs.At(1, 1) + At(1, 2) * rhs.At(2, 1);
        auto m12 = At(1, 0) * rhs.At(0, 2) + At(1, 1) * rhs.At(1, 2) + At(1, 2) * rhs.At(2, 2);

        auto m20 = At(2, 0) * rhs.At(0, 0) + At(2, 1) * rhs.At(1, 0) + At(2, 2) * rhs.At(2, 0);
        auto m21 = At(2, 0) * rhs.At(0, 1) + At(2, 1) * rhs.At(1, 1) + At(2, 2) * rhs.At(2, 1);
        auto m22 = At(2, 0) * rhs.At(0, 2) + At(2, 1) * rhs.At(1, 2) + At(2, 2) * rhs.At(2, 2);

        return Matrix3x3({m00, m01, m02}, {m10, m11, m12}, {m20, m21, m22});
    }

    Matrix3x3::Matrix3x3(float val) {
        this->elems[0][0] = val;
        this->elems[0][1] = val;
        this->elems[0][2] = val;

        this->elems[1][0] = val;
        this->elems[1][1] = val;
        this->elems[1][2] = val;

        this->elems[2][0] = val;
        this->elems[2][1] = val;
        this->elems[2][2] = val;

    }

    Matrix3x3::Matrix3x3(const Vector3 &col0, const Vector3 &col1, const Vector3 &col2) {
        SetColumn(0, col0);
        SetColumn(1, col1);
        SetColumn(2, col2);

        //this->elems[0][0] = r1.x;
        //this->elems[0][1] = r1.y;
        //this->elems[0][2] = r1.z;

        //this->elems[1][0] = r2.x;
        //this->elems[1][1] = r2.y;
        //this->elems[1][2] = r2.z;

        //this->elems[2][0] = r3.x;
        //this->elems[2][1] = r3.y;
        //this->elems[2][2] = r3.z;
    }

    Matrix3x3::Matrix3x3(const Quaternion &orientation) {
        SetRotatePart(orientation);
    }

    Matrix3x3::Matrix3x3(const EulerAngle &orientation) {
        auto sa = std::sin(orientation.pitch);
        auto ca = std::cos(orientation.pitch);
        auto sb = std::sin(orientation.roll);
        auto cb = std::cos(orientation.roll);
        auto sh = std::sin(orientation.yaw);
        auto ch = std::cos(orientation.yaw);

        At(0, 0) = ch*ca;
        At(0, 1) = -ch*sa*cb + sh*sh;
        At(0, 2) = ch*sa*sb + sh*cb;
        At(1, 0) = sa;
        At(1, 1) = ca*cb;
        At(1, 2) = -ca*cb;
        At(2, 0) = -sh*ca;
        At(2, 1) = sh*sa*cb + ch*sb;
        At(2, 2) = -sh*sa*sb + ch*cb;
    }

    float Matrix3x3::Determinant() const {
        const float a = elems[0][0];
        const float b = elems[0][1];
        const float c = elems[0][2];
        const float d = elems[1][0];
        const float e = elems[1][1];
        const float f = elems[1][2];
        const float g = elems[2][0];
        const float h = elems[2][1];
        const float i = elems[2][2];

        // Aight, what the fuck?
        return a*(e*i - f*h) + b*(f*g - d*i) + c*(d*h - e*g);
    }

    Matrix3x3 Matrix3x3::Inverted() const {
        // Compute the inverse directly using Cramer's rule
        // Warning: This method is numerically very unstable!
        float d = Determinant();

        d = 1.f / d;

        Matrix3x3 i = {
                d * (At(1, 1) * At(2, 2) - At(1, 2) * At(2, 1)),
                d * (At(0, 2) * At(2, 1) - At(0, 1) * At(2, 2)),
                d * (At(0, 1) * At(1, 2) - At(0, 2) * At(1, 1)),

                d * (At(1, 2) * At(2, 0) - At(1, 0) * At(2, 2)),
                d * (At(0, 0) * At(2, 2) - At(0, 2) * At(2, 0)),
                d * (At(0, 2) * At(1, 0) - At(0, 0) * At(1, 2)),

                d * (At(1, 0) * At(2, 1) - At(1, 1) * At(2, 0)),
                d * (At(2, 0) * At(0, 1) - At(0, 0) * At(2,1)),
                d * (At(0,0) * At(1, 1) - At(0, 1) * At(1, 0))
        };

        return i;
    }

    Matrix3x3 Matrix3x3::Transposed() const {
        auto m00 = this->elems[0][0];
        auto m01 = this->elems[0][1];
        auto m02 = this->elems[0][2];

        auto m10 = this->elems[1][0];
        auto m11 = this->elems[1][1];
        auto m12 = this->elems[1][2];

        auto m20 = this->elems[2][0];
        auto m21 = this->elems[2][1];
        auto m22 = this->elems[2][2];

        // NO: This is correct order for transposition!
        return {
                m00, m10, m20,
                m01, m11, m21,
                m02, m12, m22
        };
    }

    Vector2 Matrix3x3::Transform(const Vector2 &rhs) const {
        return {
                At(0,0) * rhs.x + At(0, 1) * rhs.y,
                At(1,0) * rhs.x + At(1, 1) * rhs.y
        };
    }

    Vector3 Matrix3x3::Transform(const Vector3 &rhs) const {
        return {
                At(0, 0) * rhs.x + At(0, 1) * rhs.y + At(0, 2) * rhs.z,
                At(1, 0) * rhs.x + At(1, 1) * rhs.y + At(1, 2) * rhs.z,
                At(2, 0) * rhs.x + At(2, 1) * rhs.y + At(2, 2) * rhs.z
        };
    }

    Quaternion Matrix3x3::ToQuat() const {
        auto m00 = At(0,0);
        auto m01 = At(0, 1);
        auto m02 = At(0, 2);
        auto m10 = At(1,0);
        auto m11 = At(1, 1);
        auto m12 = At(1, 2);
        auto m20 = At(2,0);
        auto m21 = At(2, 1);
        auto m22 = At(2, 2);

        auto w = std::sqrt(1.f + m00 + m11 + m22) / 2.f;
        float w4 = (4.f * w);
        return {
                (m21 - m12) / w4,
                (m02 - m20) / w4,
                (m10 - m01) / w4,
                w
        };
    }

    void Matrix3x3::SetRotatePart(const Vector3 &a, float angle) {
        float s = std::sin(angle);
        float c = std::cos(angle);

        const float c1 = 1.f - c;

        elems[0][0] = c+c1*a.x*a.x;
        elems[1][0] = c1*a.x*a.y+s*a.z;
        elems[2][0] = c1*a.x*a.z-s*a.y;

        elems[0][1] = c1*a.x*a.y-s*a.z;
        elems[1][1] = c+c1*a.y*a.y;
        elems[2][1] = c1*a.y*a.z+s*a.x;

        elems[0][2] = c1*a.x*a.z+s*a.y;
        elems[1][2] = c1*a.y*a.z-s*a.x;
        elems[2][2] = c+c1*a.z*a.z;
    }

    Matrix3x3 Matrix3x3::RotateAxisAngle(const Vector3 &axis, float angleRadians) {
        Matrix3x3 r;
        r.SetRotatePart(axis, angleRadians);
        return r;
    }

    void Matrix3x3::SetRow(int i, const Vector3 &vec) {
        elems[i][0] = vec.x;
        elems[i][1] = vec.y;
        elems[i][2] = vec.z;
    }

    void Matrix3x3::SetColumn(int i, const Vector3& vec)
    {
        elems[0][i] = vec.x;
        elems[1][i] = vec.y;
        elems[2][i] = vec.z;
    }

    Matrix3x3
    Matrix3x3::LookAt(const Vector3 &forward, const Vector3 &target, const Vector3 &localUp, const Vector3 &worldUp) {
        // User must input proper normalized input direction vectors
        // In the local space, the forward and up directions must be perpendicular to be well-formed.
        // In the world space, the target direction and world up cannot be degenerate (co-linear)
        // Generate the third basis vector in the local space;
        Vector3 localRight = localUp.Cross(forward).Normalized();
        // A. Now we have an orthonormal linear basis {localRight, localUp, forward} for the object local space.
        // Generate the third basis vector for the world space
        Vector3 worldRight = worldUp.Cross(target).Normalized();
        // Since the input worldUp vector is not necessarily perpendicular to the target direction vector
        // We need to compute the real world space up vector that the "head" of the object will point
        // towards when the model is looking towards the desired target direction
        Vector3 perpWorldUp = target.Cross(worldRight).Normalized();
        // B. Now we have an orthonormal linear basis {worldRight, perpWorldUp, targetDirection } for the desired target orientation.
        // We want to build a matrix M that performs the following mapping:
        // 1. localRight must be mapped to worldRight.        (M * localRight = worldRight)
        // 2. localUp must be mapped to perpWorldUp.          (M * localUp = perpWorldUp)
        // 3. localForward must be mapped to targetDirection. (M * localForward = targetDirection)
        // i.e. we want to map the basis A to basis B.

        // This matrix M exists, and it is an orthonormal rotation matrix with a determinant of +1, because
        // the bases A and B are orthonormal with the same handedness.

        // Below, use the notation that (a,b,c) is a 3x3 matrix with a as its first column, b second, and c third.

        // By algebraic manipulation, we can rewrite conditions 1, 2 and 3 in a matrix form:
        //        M * (localRight, localUp, localForward) = (worldRight, perpWorldUp, targetDirection)
        // or     M = (worldRight, perpWorldUp, targetDirection) * (localRight, localUp, localForward)^{-1}.
        // or     M = m1 * m2, where

        // m1 equals (worldRight, perpWorldUp, target):
        Matrix3x3 m1(worldRight, perpWorldUp, target);

        // and m2 equals (localRight, localUp, localForward)^{-1}:
        Matrix3x3 m2;
        m2.SetRow(0, localRight);
        m2.SetRow(1, localUp);
        m2.SetRow(2, forward);
        // Above we used the shortcut that for an orthonormal matrix M, M^{-1} = M^T. So set the rows
        // and not the columns to directly produce the transpose, i.e. the inverse of (localRight, localUp, localForward).

        // Compute final M.
        m2 = m1 * m2;

        // And fix any numeric stability issues by re-orthonormalizing the result.
        m2.Orthonormalize(0, 1, 2);
        return m2;
    }

    Vector3 Matrix3x3::Diagonal() const {
        return {elems[0][0],
                elems[1][1],
                elems[2][2]
        };
    }

    float &Matrix3x3::At(int row, int col) {
        return elems[row][col];
    }

    Vector3 Matrix3x3::WorldZ() const {
        return GetColumn(2);
    }

    Vector3 Matrix3x3::WorldY() const {
        return GetColumn(1);
    }

    Vector3 Matrix3x3::WorldX() const {
        return GetColumn(0);
    }

    Matrix3x3 Matrix3x3::FromRS(const Quaternion &rotate, const Vector3 &scale) {
        return Matrix3x3(rotate) * Matrix3x3::FromScale(scale);
    }

    Matrix3x3 Matrix3x3::FromRS(const Matrix3x3 &rotate, const Vector3 &scale) {
        return rotate * Matrix3x3::FromScale(scale);
    }

    Matrix3x3 Matrix3x3::FromQuat(const Quaternion &orientation) {
        return Matrix3x3(orientation);
    }

    Matrix3x3 Matrix3x3::ScaleBy(const Vector3 &rhs) {
        return *this * FromScale(rhs);
    }

    Vector3 Matrix3x3::GetScale() const {
        return Vector3(GetColumn(0).Length(), GetColumn(1).Length(), GetColumn(2).Length());
    }

    Vector3 Matrix3x3::operator[](int row) const {
        return Vector3{elems[row][0], elems[row][1], elems[row][2]};
    }

    Matrix3x3 Matrix3x3::FromScale(const Vector3 &scale) {
        Matrix3x3 m;
        m.At(0,0) = scale.x;
        m.At(1,1) = scale.y;
        m.At(2,2) = scale.z;
        return m;
    }

    Matrix3x3 Matrix3x3::FromScale(float sx, float sy, float sz) {
        Matrix3x3 m;
        m.At(0,0) = sx;
        m.At(1,1) = sy;
        m.At(2,2) = sz;
        return m;
    }

    void Matrix3x3::Orthonormalize(int c0, int c1, int c2) {
        Vector3 v0 = GetColumn(c0);
        Vector3 v1 = GetColumn(c1);
        Vector3 v2 = GetColumn(c2);
        Vector3::Orthonormalize(v0, v1, v2);
        SetColumn(c0, v0);
        SetColumn(c1, v1);
        SetColumn(c2, v2);
    }

    Matrix3x3::Matrix3x3(const float *data) {
        assert(data);
        At(0, 0) = data[0];
        At(0, 1) = data[1];
        At(0, 2) = data[2];

        At(1, 0) = data[4];
        At(1, 1) = data[5];
        At(1, 2) = data[6];

        At(2, 0) = data[8];
        At(2, 1) = data[9];
        At(2, 2) = data[10];
    }

    Vector4 Matrix3x3::Mul(const Vector4 &rhs) const {
        return Vector4(Mul(rhs.XYZ()), rhs.GetW());
    }

    Vector3 Matrix3x3::Mul(const Vector3 &rhs) const {
        return *this * rhs;
    }

    Vector2 Matrix3x3::Mul(const Vector2 &rhs) const {
        return *this * rhs;
    }

    Matrix4x4 Matrix3x3::Mul(const Matrix4x4 &rhs) const {
        return *this * rhs;
    }

    Matrix3x3 Matrix3x3::Mul(const Matrix3x3 &rhs) const {
        return *this * rhs;
    }

    bool Matrix3x3::IsRowOrthogonal(float epsilon) const
    {
        return GetRow(0).IsPerpendicular(GetRow(1), epsilon)
            && GetRow(0).IsPerpendicular(GetRow(2), epsilon)
            && GetRow(1).IsPerpendicular(GetRow(2), epsilon);
    }

    bool Matrix3x3::IsColOrthogonal(float epsilon) const
    {
        return GetColumn(0).IsPerpendicular(GetColumn(1), epsilon)
            && GetColumn(0).IsPerpendicular(GetColumn(2), epsilon)
            && GetColumn(1).IsPerpendicular(GetColumn(2), epsilon);
    }



    bool Matrix3x3::HasUniformScale(float epsilon) const {
        Vector3 scale = ExtractScale();
        return Math::EqualAbs(scale.x, scale.y, epsilon) && Math::EqualAbs(scale.x, scale.z, epsilon);
    }

    Vector3 Matrix3x3::GetRow3(int index) const {
        return GetRow(index);
    }

    Vector3 Matrix3x3::GetColumn3(int index) const {
        return GetColumn(index);
    }

    Matrix4x4 Matrix3x3::operator*(const Matrix4x4 &rhs) const {
        auto lhs = *this;
        Matrix4x4 r;

        r[0][0] = lhs.At(0, 0) * rhs.At(0, 0) + lhs.At(0, 1) * rhs.At(1, 0) + lhs.At(0, 2) * rhs.At(2, 0);
        r[0][1] = lhs.At(0, 0) * rhs.At(0, 1) + lhs.At(0, 1) * rhs.At(1, 1) + lhs.At(0, 2) * rhs.At(2, 1);
        r[0][2] = lhs.At(0, 0) * rhs.At(0, 2) + lhs.At(0, 1) * rhs.At(1, 2) + lhs.At(0, 2) * rhs.At(2, 2);
        r[0][3] = lhs.At(0, 0) * rhs.At(0, 3) + lhs.At(0, 1) * rhs.At(1, 3) + lhs.At(0, 2) * rhs.At(2, 3);

        r[1][0] = lhs.At(1, 0) * rhs.At(0, 0) + lhs.At(1, 1) * rhs.At(1, 0) + lhs.At(1, 2) * rhs.At(2, 0);
        r[1][1] = lhs.At(1, 0) * rhs.At(0, 1) + lhs.At(1, 1) * rhs.At(1, 1) + lhs.At(1, 2) * rhs.At(2, 1);
        r[1][2] = lhs.At(1, 0) * rhs.At(0, 2) + lhs.At(1, 1) * rhs.At(1, 2) + lhs.At(1, 2) * rhs.At(2, 2);
        r[1][3] = lhs.At(1, 0) * rhs.At(0, 3) + lhs.At(1, 1) * rhs.At(1, 3) + lhs.At(1, 2) * rhs.At(2, 3);

        r[2][0] = lhs.At(2, 0) * rhs.At(0, 0) + lhs.At(2, 1) * rhs.At(1, 0) + lhs.At(2, 2) * rhs.At(2, 0);
        r[2][1] = lhs.At(2, 0) * rhs.At(0, 1) + lhs.At(2, 1) * rhs.At(1, 1) + lhs.At(2, 2) * rhs.At(2, 1);
        r[2][2] = lhs.At(2, 0) * rhs.At(0, 2) + lhs.At(2, 1) * rhs.At(1, 2) + lhs.At(2, 2) * rhs.At(2, 2);
        r[2][3] = lhs.At(2, 0) * rhs.At(0, 3) + lhs.At(2, 1) * rhs.At(1, 3) + lhs.At(2, 2) * rhs.At(2, 3);

        r[3][0] = rhs.At(3, 0);
        r[3][1] = rhs.At(3, 1);
        r[3][2] = rhs.At(3, 2);
        r[3][3] = rhs.At(3, 3);
        return r;
    }

    Vector2 Matrix3x3::operator*(const Vector2 &rhs) const {
        return Transform(rhs);
    }

    Matrix3x3 Matrix3x3::RotateX(float radians) {
        Matrix3x3 r;
        r.SetRotatePartX(radians);
        return r;
    }

    Matrix3x3 Matrix3x3::RotateY(float radians) {
        Matrix3x3 r;
        r.SetRotatePartY(radians);
        return r;
    }

    Matrix3x3 Matrix3x3::RotateZ(float radians) {
        Matrix3x3 r;
        r.SetRotatePartZ(radians);
        return r;
    }

    void Matrix3x3::SetRotatePartX(float angle) {
        Set3x3PartRotateX(*this, angle);
    }

    void Matrix3x3::SetRotatePartY(float angle) {
        Set3x3PartRotateY(*this, angle);
    }

    void Matrix3x3::SetRotatePartZ(float angle) {
        Set3x3PartRotateZ(*this, angle);
    }

    Vector3 Matrix3x3::ExtractScale() const {
        return {GetColumn(0).Length(), GetColumn(1).Length(), GetColumn(2).Length()};
    }

    // TODO: Finish implementation
    Matrix3x3 Matrix3x3::RotateFromTo(const Vector3 &source, const Vector3 &direction) {
        assert(source.IsNormalized());
        assert(source.IsNormalized());

        // http://cs.brown.edu/research/pubs/pdfs/1999/Moller-1999-EBA.pdf
        Matrix3x3 r;
        float dot = source.Dot(direction);
        if (std::abs(dot) > 0.999f)
        {
            Vector3 s = source.Abs();
            Vector3 unit = s.x < s.y && s.x < s.z ? Vector3::UnitX : (s.y < s.z ? Vector3::UnitY : Vector3::UnitZ);
        }

        return Matrix3x3::Identity;
    }

    Vector3 &Matrix3x3::Row(int row) {
        assert(row >= 0);
        assert(row < Rows);
        return reinterpret_cast<Vector3 &> (elems[row]);
    }

    Vector3 Matrix3x3::Column(int index) const { return GetColumn(index);}

    Vector3 Matrix3x3::Col(int index) const { return Column(index);}

    Vector3 &Matrix3x3::Row3(int index) {
        return reinterpret_cast<Vector3 &>(elems[index]);
    }

    Vector3 Matrix3x3::Row3(int index) const { return GetRow3(index);}

    void Matrix3x3::Set(const Matrix3x3 &rhs) {
        Set(rhs.ptr());
    }

    bool Matrix3x3::IsFinite() const {
        for (int y = 0; y < Rows; y++)
            for (int x = 0; x < Cols; ++x)
                if (!std::isfinite(elems[y][x]))
                    return false;
        return true;
    }

    /** Compares the two values for equality, allowing the given amount of absolute error. */
    bool EqualAbs(float a, float b, float epsilon)
    {
        return std::abs(a-b) < epsilon;
    }

    bool Matrix3x3::IsIdentity(float epsilon) const
    {
        for (int y = 0; y < Rows; ++y)
            for (int x = 0; x < Cols; ++x)
                if (!EqualAbs(elems[y][x], (x == y) ? 1.f : 0.f, epsilon))
                    return false;
        return true;
    }


    bool Matrix3x3::IsLowerTriangular(float epsilon) const
    {
        return EqualAbs(elems[0][1], 0.f, epsilon)
            && EqualAbs(elems[0][2], 0.f, epsilon)
            && EqualAbs(elems[1][2], 0.f, epsilon);
    }


    bool Matrix3x3::IsUpperTriangular(float epsilon) const
    {
        return EqualAbs(elems[1][0], 0.f, epsilon)
            && EqualAbs(elems[2][0], 0.f, epsilon)
            && EqualAbs(elems[2][1], 0.f, epsilon);
    }

    bool Matrix3x3::IsInvertible(float epsilon) const
    {
        float d = Determinant();
        bool isSingular = EqualAbs(d, 0.f, epsilon);
        //assert(Inverted(epsilon) == isSingular);
        return !isSingular;
    }

    bool Matrix3x3::IsSymmetric(float epsilon) const
    {
        return EqualAbs(elems[0][1], elems[1][0], epsilon)
            && EqualAbs(elems[0][2], elems[2][0], epsilon)
            && EqualAbs(elems[1][2], elems[2][1], epsilon);
    }

    bool Matrix3x3::IsSkewSymmetric(float epsilon) const
    {
        return EqualAbs(elems[0][0], 0.f, epsilon)
            && EqualAbs(elems[1][1], 0.f, epsilon)
            && EqualAbs(elems[2][2], 0.f, epsilon)
            && EqualAbs(elems[0][1], -elems[1][0], epsilon)
            && EqualAbs(elems[0][2], -elems[2][0], epsilon)
            && EqualAbs(elems[1][2], -elems[2][1], epsilon);
    }


    bool Matrix3x3::HasUnitaryScale(float epsilon) const {
        Vector3 scale = ExtractScale();
        return scale.Equals(1.f, 1.f, 1.f, epsilon);
    }

    bool Matrix3x3::HasNegativeScale() const
    {
        return Determinant() < 0.f;
    }


    bool Matrix3x3::IsOrthonormal(float epsilon) const
    {
        ///@todo Epsilon magnitudes don't match.
        return IsColOrthogonal(epsilon) && Row(0).IsNormalized(epsilon) && Row(1).IsNormalized(epsilon) && Row(2).IsNormalized(epsilon);
    }

    Vector3 Matrix3x3::Row(int index) const { return GetRow(index);}

    float Matrix3x3::DeterminantSymmetric() const {
        assert(IsSymmetric());

        const float a = this->elems[0][0];
        const float b = this->elems[0][1];
        const float c = this->elems[0][2];
        const float d = this->elems[1][0];
        const float e = this->elems[1][1];
        const float f = this->elems[1][2];

        // If the matrix is symmetric, it is of the following form:
        // a b c
        // b d e
        // c e f

        // A direct cofactor expansion gives
        // det = a * (df - ee) -b * (bf - ce) + c * (be-dc)
        //     = adf - aee - bbf + bce + bce - ccd
        //     = adf - aee - bbf - ccd + 2*bce
        //     = a(df-ee) + b(2*ce - bf) - ccd

        return a * (d*f - e*e) + b * (2.f * c * e - b * f) - c*c*d;
    }

    Matrix3x3 Matrix3x3::InverseTransposed() const {
        Matrix3x3 copy = *this;
        copy.Transpose();
        copy.Inverse();
        return copy;
    }

    bool Matrix3x3::InverseTranspose() {
        bool success = Inverse();
        Transpose();
        return success;
    }

    bool Matrix3x3::Inverse(float epsilon) {
        /// There exists a generic matrix inverse calculator that uses Gaussian elimination

        Matrix3x3 i = *this;
        bool success = InverseMatrix(i, epsilon);
        if (!success)
            return false;

        *this = i;
        return true;
    }

    bool Matrix3x3::InverseFast(float epsilon) {
        // Compute the inverse directly using Cramer's rule.
        // Warning: This method is numerically very unstable!
        float d = Determinant();
        if (Math::EqualAbs(d, 0.f, epsilon))
            return true;

        d = 1.f / d;
        Matrix3x3 i;
        i[0][0] = d * (At(1, 1) * At(2, 2) - At(1, 2) * At(2, 1));
        i[0][1] = d * (At(0, 2) * At(2, 1) - At(0, 1) * At(2, 2));
        i[0][2] = d * (At(0, 1) * At(1, 2) - At(0, 2) * At(1, 1));

        i[1][0] = d * (At(1, 2) * At(2, 0) - At(1, 0) * At(2, 2));
        i[1][1] = d * (At(0, 0) * At(2, 2) - At(0, 2) * At(2, 0));
        i[1][2] = d * (At(0, 2) * At(1, 0) - At(0, 0) * At(1, 2));

        i[2][0] = d * (At(1, 0) * At(2, 1) - At(1, 1) * At(2, 0));
        i[2][1] = d * (At(2, 0) * At(0, 1) - At(0, 0) * At(2, 1));
        i[2][2] = d * (At(0, 0) * At(1, 1) - At(0, 1) * At(1, 0));


        *this = i;
        return true;
    }

    bool Matrix3x3::SolveAxb(Vector3 b, Vector3 &x) const {
        // Solve by pivotization.
        float v00 = At(0, 0);
        float v10 = At(1, 0);
        float v20 = At(2, 0);

        float v01 = At(0, 1);
        float v11 = At(1, 1);
        float v21 = At(2, 1);

        float v02 = At(0, 2);
        float v12 = At(1, 2);
        float v22 = At(2, 2);

        float av00 = std::abs(v00);
        float av10 = std::abs(v10);
        float av20 = std::abs(v20);

        // Find which item in first column has largest absolute value.
        if (av10 >= av00 && av10 >= av20) {
            Swap(v00, v10);
            Swap(v01, v11);
            Swap(v02, v12);
            Swap(b[0], b[1]);
        } else if (av20 >= av00) {
            Swap(v00, v20);
            Swap(v01, v21);
            Swap(v02, v22);
            Swap(b[0], b[2]);
        }

        /* a b c | x
           d e f | y
           g h i | z, where |a| >= |d| && |a| >= |g| */

        if (Math::EqualAbs(v00, 0.f))
            return false;


        // Scale row so that leading element is one.
        float denom = 1.f / v00;
        v01 *= denom;
        v02 *= denom;
        b[0] *= denom;

        /* 1 b c | x
           d e f | y
           g h i | z */


        // Pivotize again.
        if (std::abs(v21) >= std::abs(v11))
        {
            Swap(v11, v21);
            Swap(v12, v22);
            Swap(b[1], b[2]);
        }

        if (Math::EqualAbs(v11, 0.f))
            return false;

        /* 1 b c | x
           0 e f | y
           0 h i | z, where |e| >= |h| */

        denom = 1.f / v11;
        // v11 = 1.f;
        v12 *= denom;
        b[1] *= denom;

        /* 1 b c | x
           0 1 f | y
           0 h i | z */

        if (Math::EqualAbs(v22, 0.f))
            return false;

        x[2] = b[2] / v22;
        x[1] = b[1] - x[2] * v12;
        x[0] = b[0] - x[2] * v02 - x[1] * v01;



        return true;
    }

    void Matrix3x3::ScaleCol(int col, float scalar) {
        assert(col >= 0);
        assert(col < Cols);
        At(0, col) *= scalar;
        At(1, col) *= scalar;
        At(2, col) *= scalar;
    }

    void Matrix3x3::ScaleRow(int row, float scalar) {
        assert(row >= 0);
        assert(row < Rows);
        assert(std::isfinite(scalar));
        At(row, 0) *= scalar;
        At(row, 1) *= scalar;
        At(row, 2) *= scalar;
    }

    Vector3 Matrix3x3::Column3(int index) const {
        return GetColumn3(index);
    }

    Vector3 Matrix3x3::Col3(int index) const {
        return GetColumn3(index);
    }

    void Matrix3x3::Transpose() {
        Swap(elems[0][1], elems[1][0]);
        Swap(elems[0][2], elems[2][0]);
        Swap(elems[1][2], elems[2][1]);
    }

    void Matrix3x3::SwapRows(int row1, int row2) {
        assert(row1 >= 0);
        assert(row1 < Rows);
        assert(row2 >= 0);
        assert(row2 < Rows);
        Swap(At(row1, 0), At(row2, 0));
        Swap(At(row1, 1), At(row2, 1));
        Swap(At(row1, 2), At(row2, 2));
    }

    void Matrix3x3::SetIdentity() {
        Set(1,0,0,
            0,1,0,
            0,0,1);
    }

    void Matrix3x3::SwapColumns(int col1, int col2) {
        assert(col1 >= 0);
        assert(col1 < Cols);
        assert(col2 >= 0);
        assert(col2 < Cols);
        Swap(At(0, col1), At(0, col2));
        Swap(At(1, col1), At(1, col2));
        Swap(At(2, col1), At(2, col2));
    }

    void Matrix3x3::Set(const float *p) {
        assert(p);
        Set(p[0], p[1], p[2],
            p[3], p[4], p[5],
            p[6], p[7], p[8]);
    }

    void
    Matrix3x3::Set(float _00, float _01, float _02, float _10, float _11, float _12, float _20, float _21, float _22) {
        At(0, 0) = _00; At(0, 1) = _01; At(0, 2) = _02;
        At(1, 0) = _10; At(1, 1) = _11; At(1, 2) = _12;
        At(2, 0) = _20; At(2, 1) = _21; At(2, 2) = _22;
    }

    void Matrix3x3::SetRotatePart(const Quaternion &quat) {
        SetMatrixRotatePart(*this, quat);
    }

    void Matrix3x3::RemoveScale() {
        float x = Row(0).Normalize();
        float y = Row(1).Normalize();
        float z = Row(2).Normalize();

        assert(x != 0 && "Matrix3x3::RemoveScale failed!");
        assert(y != 0 && "Matrix3x3::RemoveScale failed!");
        assert(z != 0 && "Matrix3x3::RemoveScale failed!");
    }

    float Matrix3x3::Trace() const {
        return elems[0][0] + elems[1][1] + elems[2][2];
    }

    bool Matrix3x3::InverseSymmetric() {
        assert(IsSymmetric());
        const float a = elems[0][0];
        const float b = elems[0][1];
        const float c = elems[0][2];
        const float d = elems[1][1];
        const float e = elems[1][2];
        const float f = elems[2][2];

        /* If the matrix is symmetric, it is of form
           a b c
           b d e
           c e f
         */

        // A direct cofactor expansion gives
        // det = a * (df - ee) + b * (ce - bf) + c * (be-dc)


        const float df_ee = d*f - e*e;
        const float ce_bf = c*e - b*f;
        const float be_dc = b*e - d*c;

        float det = a * df_ee + b * ce_bf + c * be_dc;
        if (Math::EqualAbs(det, 0.f))
            return false;

        det = 1.f / det;


        // The inverse of a symmetric matrix will also be symmetric, so we can avoid some computations altogether.

        elems[0][0] = det * df_ee;
        elems[1][0] = elems[0][1] = det * ce_bf;
        elems[0][2] = elems[2][0] = det * be_dc;
        elems[1][1] = det * (a*f - c*c);
        elems[1][2] = elems[2][1] = det * (c*b - a*e);
        elems[2][2] = det * (a*d - b*b);

        return true;
    }

    void Matrix3x3::InverseOrthonormal() {
        Matrix3x3 orig = *this;

        assert(IsOrthonormal());

        Transpose();
        assert(!orig.IsInvertible() || (orig * *this).IsIdentity());
    }

    bool Matrix3x3::InverseOrthogonalUniformScale() {
        Matrix3x3 orig = *this;
        assert(IsColOrthogonal());
        assert(HasUniformScale());
        float scale = Vector3(At(0, 0), At(1, 0), At(2, 0)).LengthSquared();
        if (scale < 1e-8f)
            return false;
        scale = 1.f / scale;
        Swap(At(0, 1), At(1, 0));
        Swap(At(0, 2), At(2, 0));
        Swap(At(1, 2), At(2, 1));

        At(0, 0) *= scale; At(0, 1) *= scale; At(0, 2) *= scale;
        At(1, 0) *= scale; At(1, 1) *= scale; At(1, 2) *= scale;
        At(2, 0) *= scale; At(2, 1) *= scale; At(2, 2) *= scale;

        assert(IsFinite());
        assert(IsColOrthogonal());
        assert(HasUniformScale());

        assert(!orig.IsInvertible() || (orig * *this).IsIdentity());
        return true;
    }

    bool Matrix3x3::InverseColOrthogonal() {
        Matrix3x3 orig = *this;
        assert(IsColOrthogonal());
        float s1 = Vector3(At(0, 0), At(1, 0), At(2, 0)).LengthSquared();
        float s2 = Vector3(At(0, 1), At(1, 1), At(2, 1)).LengthSquared();
        float s3 = Vector3(At(0, 2), At(1, 2), At(2, 2)).LengthSquared();
        if (s1 < 1e-8f || s2 < 1e-8f || s3 < 1e-8f)
            return false;

        s1 = 1.f / s1;
        s2 = 1.f / s2;
        s3 = 1.f / s3;

        Swap(At(0, 1), At(1, 0));
        Swap(At(0, 2), At(2, 0));
        Swap(At(1, 2), At(2, 1));

        At(0, 0) *= s1; At(0, 1) *= s1; At(0, 2) *= s1;
        At(1, 0) *= s2; At(1, 1) *= s2; At(1, 2) *= s2;
        At(2, 0) *= s3; At(2, 1) *= s3; At(2, 2) *= s3;

        assert(!orig.IsInvertible() || (orig * *this).IsIdentity());
        assert(IsRowOrthogonal());

        return true;
    }

    bool Matrix3x3::TryConvertToQuat(Quaternion &q) const {
        if (IsColOrthogonal() && HasUnitaryScale() && !HasNegativeScale()) {
            q = ToQuat();
            return true;
        }
        return false;
    }

    EulerAngle Matrix3x3::ToEulerAngle() const {
        auto heading = std::atan2(-At(2, 0), At(0, 0));
        auto attitude = std::asin(At(1, 0));
        auto bank = std::atan2(-At(1,2), At(1,1));
        if (At(1, 0) == 1 || At(1, 0) == -1) // North Pole || South Pole
        {
            heading = std::atan2(At(0, 2), At(2,2));
            bank = 0;
        }

        return {attitude, heading, bank};
    }

    void Matrix3x3::BatchTransform(Vector3 *pointArray, int numPoints, int stride) const {
        assert(pointArray || numPoints == 0);
        assert(stride >= (int)sizeof(Vector3));
        u8 *data = reinterpret_cast<u8*>(pointArray);

        for (int i = 0; i < numPoints; ++i)
        {
            Vector3 *vtx = reinterpret_cast<Vector3*>(data + stride*i);
            *vtx = *this * *vtx;
        }
    }

    void Matrix3x3::BatchTransform(Vector3 *pointArray, int numPoints) const {
        assert(pointArray || numPoints == 0);

        for (int i = 0; i < numPoints; ++i)
        {
            pointArray[i] = *this * pointArray[i];
        }
    }

    void Matrix3x3::BatchTransform(Vector4 *vectorArray, int numVectors) const {
        assert(vectorArray || numVectors == 0);
        for(int i = 0; i < numVectors; ++i)
            vectorArray[i] = *this * vectorArray[i];
    }

    void Matrix3x3::BatchTransform(Vector4 *vectorArray, int numVectors, int stride) const {
        assert(vectorArray || numVectors == 0);
        assert(stride >= (int)sizeof(Vector4));
        u8 *data = reinterpret_cast<u8*>(vectorArray);
        for(int i = 0; i < numVectors; ++i)
        {
            Vector4 *vtx = reinterpret_cast<Vector4*>(data + stride*i);
            *vtx = *this * *vtx;
        }
    }

    Vector4 Matrix3x3::operator*(const Vector4 &rhs) const {
        return Vector4();
    }

    void Matrix3x3::SetRow(int row, const float *data) {
        assert(data);
        SetRow(row, data[0], data[1], data[2]);
    }

    void Matrix3x3::SetRow(int row, float x, float y, float z) {
        assert(row > 0);
        assert(row < Rows);
        assert(std::isfinite(x));
        assert(std::isfinite(y));
        assert(std::isfinite(z));
        At(row, 0) = x;
        At(row, 1) = y;
        At(row, 2) = z;
    }

    void Matrix3x3::SetColumn(int column, const float *data) {
        assert(data);
        SetColumn(column, data[0], data[1], data[2]);
    }

    void Matrix3x3::SetColumn(int column, float x, float y, float z) {
        assert(column >= 0);
        assert(column < Cols);
        assert(std::isfinite(x));
        assert(std::isfinite(y));
        assert(std::isfinite(z));
        At(0, column) = x;
        At(1, column) = y;
        At(2, column) = z;
    }

    Matrix3x3 Matrix3x3::Mul(const Quaternion &rhs) const {
        return *this * rhs;
    }

    Matrix3x3 Matrix3x3::operator*(const Quaternion &rhs) const {
        return *this * Matrix3x3(rhs);
    }

    bool Matrix3x3::Equals(const Matrix3x3 &other, float epsilon) const {
        for(int y = 0; y < Rows; ++y)
            for(int x = 0; x < Cols; ++x)
                if (!Math::EqualAbs(At(y, x), other[y][x], epsilon))
                    return false;
        return true;
    }

    Matrix3x3 Matrix3x3::RandomRotation(RNG &rng) {
        // The easiest way to generate a random orientation is through quaternions
        // So we convert a random quaternion to a rotation matrix.
        return FromQuat(Quaternion::RandomRotation(rng));
    }

    Matrix3x3 Matrix3x3::RandomGeneral(RNG &rng, float minElem, float maxElem) {
        Matrix3x3 m;
        for (int y = 0; y < 3; ++y)
        {
            for (int x = 0; x < 3; ++x)
            {
                m[y][x] = rng.Float(minElem, maxElem);
            }
        }
        return m;
    }


}

